# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["LinodeDNSDriver"]

from datetime import datetime

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import Provider, RecordType, ZoneDoesNotExistError, RecordDoesNotExistError
from libcloud.utils.py3 import httplib
from libcloud.utils.misc import get_new_obj, merge_valid_keys
from libcloud.common.linode import (
    API_ROOT,
    DEFAULT_API_VERSION,
    LinodeResponse,
    LinodeException,
    LinodeConnection,
    LinodeResponseV4,
    LinodeExceptionV4,
    LinodeConnectionV4,
)

try:
    import simplejson as json
except ImportError:
    import json


VALID_ZONE_EXTRA_PARAMS = [
    "SOA_Email",
    "Refresh_sec",
    "Retry_sec",
    "Expire_sec",
    "status",
    "master_ips",
]

VALID_RECORD_EXTRA_PARAMS = ["Priority", "Weight", "Port", "Protocol", "TTL_sec"]

VALID_ZONE_EXTRA_PARAMS_V4 = [
    "description",
    "expire_sec",
    "master_ips",
    "refresh_sec",
    "retry_sec",
    "soa_email",
    "status",
    "tags",
]

VALID_RECORD_EXTRA_PARAMS_V4 = [
    "port",
    "priority",
    "protocol",
    "service",
    "tag",
    "ttl_sec",
    "target",
    "weight",
]


class LinodeDNSDriver(DNSDriver):
    type = Provider.LINODE
    name = "Linode DNS"
    website = "http://www.linode.com/"

    def __new__(
        cls,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=DEFAULT_API_VERSION,
        **kwargs,
    ):
        if cls is LinodeDNSDriver:
            if api_version == "3.0":
                cls = LinodeDNSDriverV3
            elif api_version == "4.0":
                cls = LinodeDNSDriverV4
            else:
                raise NotImplementedError(
                    "No Linode driver found for API version: %s" % (api_version)
                )
        return super().__new__(cls)


class LinodeDNSResponse(LinodeResponse):
    def _make_excp(self, error):
        result = super()._make_excp(error)
        if isinstance(result, LinodeException) and result.code == 5:
            context = self.connection.context

            if context["resource"] == "zone":
                result = ZoneDoesNotExistError(
                    value="", driver=self.connection.driver, zone_id=context["id"]
                )

            elif context["resource"] == "record":
                result = RecordDoesNotExistError(
                    value="", driver=self.connection.driver, record_id=context["id"]
                )
        return result


class LinodeDNSConnection(LinodeConnection):
    responseCls = LinodeDNSResponse


class LinodeDNSDriverV3(LinodeDNSDriver):
    connectionCls = LinodeDNSConnection

    RECORD_TYPE_MAP = {
        RecordType.NS: "NS",
        RecordType.MX: "MX",
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.TXT: "TXT",
        RecordType.SRV: "SRV",
    }

    def list_zones(self):
        params = {"api_action": "domain.list"}
        data = self.connection.request(API_ROOT, params=params).objects[0]
        zones = self._to_zones(data)
        return zones

    def list_records(self, zone):
        params = {"api_action": "domain.resource.list", "DOMAINID": zone.id}

        self.connection.set_context(context={"resource": "zone", "id": zone.id})
        data = self.connection.request(API_ROOT, params=params).objects[0]
        records = self._to_records(items=data, zone=zone)
        return records

    def get_zone(self, zone_id):
        params = {"api_action": "domain.list", "DomainID": zone_id}
        self.connection.set_context(context={"resource": "zone", "id": zone_id})
        data = self.connection.request(API_ROOT, params=params).objects[0]
        zones = self._to_zones(data)

        if len(zones) != 1:
            raise ZoneDoesNotExistError(value="", driver=self, zone_id=zone_id)

        return zones[0]

    def get_record(self, zone_id, record_id):
        zone = self.get_zone(zone_id=zone_id)
        params = {
            "api_action": "domain.resource.list",
            "DomainID": zone_id,
            "ResourceID": record_id,
        }
        self.connection.set_context(context={"resource": "record", "id": record_id})
        data = self.connection.request(API_ROOT, params=params).objects[0]
        records = self._to_records(items=data, zone=zone)

        if len(records) != 1:
            raise RecordDoesNotExistError(value="", driver=self, record_id=record_id)

        return records[0]

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        Create a new zone.

        API docs: http://www.linode.com/api/dns/domain.create
        """
        params = {"api_action": "domain.create", "Type": type, "Domain": domain}

        if ttl:
            params["TTL_sec"] = ttl

        merged = merge_valid_keys(params=params, valid_keys=VALID_ZONE_EXTRA_PARAMS, extra=extra)
        data = self.connection.request(API_ROOT, params=params).objects[0]
        zone = Zone(
            id=data["DomainID"],
            domain=domain,
            type=type,
            ttl=ttl,
            extra=merged,
            driver=self,
        )
        return zone

    def update_zone(self, zone, domain=None, type=None, ttl=None, extra=None):
        """
        Update an existing zone.

        API docs: http://www.linode.com/api/dns/domain.update
        """
        params = {"api_action": "domain.update", "DomainID": zone.id}

        if type:
            params["Type"] = type

        if domain:
            params["Domain"] = domain

        if ttl:
            params["TTL_sec"] = ttl

        merged = merge_valid_keys(params=params, valid_keys=VALID_ZONE_EXTRA_PARAMS, extra=extra)
        self.connection.request(API_ROOT, params=params).objects[0]
        updated_zone = get_new_obj(
            obj=zone,
            klass=Zone,
            attributes={"domain": domain, "type": type, "ttl": ttl, "extra": merged},
        )
        return updated_zone

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a new record.

        API docs: http://www.linode.com/api/dns/domain.resource.create
        """
        params = {
            "api_action": "domain.resource.create",
            "DomainID": zone.id,
            "Name": name,
            "Target": data,
            "Type": self.RECORD_TYPE_MAP[type],
        }
        merged = merge_valid_keys(params=params, valid_keys=VALID_RECORD_EXTRA_PARAMS, extra=extra)

        result = self.connection.request(API_ROOT, params=params).objects[0]
        record = Record(
            id=result["ResourceID"],
            name=name,
            type=type,
            data=data,
            extra=merged,
            zone=zone,
            driver=self,
            ttl=merged.get("TTL_sec", None),
        )
        return record

    def update_record(self, record, name=None, type=None, data=None, extra=None):
        """
        Update an existing record.

        API docs: http://www.linode.com/api/dns/domain.resource.update
        """
        params = {
            "api_action": "domain.resource.update",
            "ResourceID": record.id,
            "DomainID": record.zone.id,
        }

        if name:
            params["Name"] = name

        if data:
            params["Target"] = data

        if type is not None:
            params["Type"] = self.RECORD_TYPE_MAP[type]

        merged = merge_valid_keys(params=params, valid_keys=VALID_RECORD_EXTRA_PARAMS, extra=extra)

        self.connection.request(API_ROOT, params=params).objects[0]
        updated_record = get_new_obj(
            obj=record,
            klass=Record,
            attributes={"name": name, "data": data, "type": type, "extra": merged},
        )
        return updated_record

    def delete_zone(self, zone):
        params = {"api_action": "domain.delete", "DomainID": zone.id}

        self.connection.set_context(context={"resource": "zone", "id": zone.id})
        data = self.connection.request(API_ROOT, params=params).objects[0]

        return "DomainID" in data

    def delete_record(self, record):
        params = {
            "api_action": "domain.resource.delete",
            "DomainID": record.zone.id,
            "ResourceID": record.id,
        }

        self.connection.set_context(context={"resource": "record", "id": record.id})
        data = self.connection.request(API_ROOT, params=params).objects[0]

        return "ResourceID" in data

    def _to_zones(self, items):
        """
        Convert a list of items to the Zone objects.
        """
        zones = []

        for item in items:
            zones.append(self._to_zone(item))

        return zones

    def _to_zone(self, item):
        """
        Build an Zone object from the item dictionary.
        """
        extra = {
            "SOA_Email": item["SOA_EMAIL"],
            "status": item["STATUS"],
            "description": item["DESCRIPTION"],
        }
        zone = Zone(
            id=item["DOMAINID"],
            domain=item["DOMAIN"],
            type=item["TYPE"],
            ttl=item["TTL_SEC"],
            driver=self,
            extra=extra,
        )
        return zone

    def _to_records(self, items, zone=None):
        """
        Convert a list of items to the Record objects.
        """
        records = []

        for item in items:
            records.append(self._to_record(item=item, zone=zone))

        return records

    def _to_record(self, item, zone=None):
        """
        Build a Record object from the item dictionary.
        """
        extra = {
            "protocol": item["PROTOCOL"],
            "ttl_sec": item["TTL_SEC"],
            "port": item["PORT"],
            "weight": item["WEIGHT"],
            "priority": item["PRIORITY"],
        }
        type = self._string_to_record_type(item["TYPE"])
        record = Record(
            id=item["RESOURCEID"],
            name=item["NAME"],
            type=type,
            data=item["TARGET"],
            zone=zone,
            driver=self,
            ttl=item["TTL_SEC"],
            extra=extra,
        )
        return record


class LinodeDNSResponseV4(LinodeResponseV4):
    pass


class LinodeDNSConnectionV4(LinodeConnectionV4):
    responseCls = LinodeDNSResponseV4


class LinodeDNSDriverV4(LinodeDNSDriver):
    connectionCls = LinodeDNSConnectionV4

    RECORD_TYPE_MAP = {
        RecordType.SOA: "SOA",
        RecordType.NS: "NS",
        RecordType.MX: "MX",
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.TXT: "TXT",
        RecordType.SRV: "SRV",
        RecordType.CAA: "CAA",
    }

    def list_zones(self):
        """
        Return a list of zones.

        :return: ``list`` of :class:`Zone`
        """
        data = self._paginated_request("/v4/domains", "data")
        return [self._to_zone(zone) for zone in data]

    def list_records(self, zone):
        """
        Return a list of records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`Record`
        """
        if not isinstance(zone, Zone):
            raise LinodeExceptionV4("Invalid zone instance")

        data = self._paginated_request("/v4/domains/%s/records" % zone.id, "data")

        return [self._to_record(record, zone=zone) for record in data]

    def get_zone(self, zone_id):
        """
        Return a Zone instance.

        :param zone_id: ID of the zone to get
        :type  zone_id: ``str``

        :rtype: :class:`Zone`
        """
        data = self.connection.request("/v4/domains/%s" % zone_id).object

        return self._to_zone(data)

    def get_record(self, zone_id, record_id):
        """
        Return a record instance.

        :param zone_id: ID of the record's zone
        :type  zone_id: ``str``

        :param record_id: ID of the record to get
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        data = self.connection.request(
            "/v4/domains/{}/records/{}".format(zone_id, record_id)
        ).object

        return self._to_record(data, self.get_zone(zone_id))

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        Create a new zone.

        :param domain: Zone domain name (e.g. example.com)
        :type domain: ``str``

        :keyword type: Zone type (master / slave).
        :type  type: ``str``

        :keyword ttl: TTL for new records. (optional)
        :type  ttl: ``int``

        :keyword extra: Extra attributes.('description', 'expire_sec', \
        'master_ips','refresh_sec', 'retry_sec', 'soa_email',\
        'status', 'tags'). 'soa_email' required for master zones
        :type extra: ``dict``

        :rtype: :class:`Zone`
        """

        attr = {
            "domain": domain,
            "type": type,
        }
        if ttl is not None:
            attr["ttl_sec"] = ttl

        merge_valid_keys(params=attr, valid_keys=VALID_ZONE_EXTRA_PARAMS_V4, extra=extra)
        data = self.connection.request("/v4/domains", data=json.dumps(attr), method="POST").object

        return self._to_zone(data)

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a record.

        :param name: The name of this Record.For A records\
        this is the subdomain being associated with an IP address.
        :type  name: ``str``

        :param zone: Zone which the records will be created for.
        :type zone: :class:`Zone`

        :param type: DNS record type.
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).\
         For A records, this is the address the domain should resolve to.
        :type  data: ``str``

        :param extra: Extra attributes.('port', 'priority', \
        'protocol', 'service', 'tag', 'ttl_sec', 'target', 'weight')
        :type  extra: ``dict``

        :rtype: :class:`Record`
        """
        if not isinstance(zone, Zone):
            raise LinodeExceptionV4("Invalid zone instance")

        attr = {
            "type": self.RECORD_TYPE_MAP[type],
            "name": name,
            "target": data,
        }

        merge_valid_keys(params=attr, valid_keys=VALID_RECORD_EXTRA_PARAMS_V4, extra=extra)
        data = self.connection.request(
            "/v4/domains/%s/records" % zone.id, data=json.dumps(attr), method="POST"
        ).object

        return self._to_record(data, zone=zone)

    def update_zone(self, zone, domain, type="master", ttl=None, extra=None):
        """
        Update an existing zone.

        :param zone: Zone to update.
        :type  zone: :class:`Zone`

        :param domain: Name of zone
        :type  domain: ``str``

        :param type: Zone type (master / slave).
        :type  type: ``str``

        :param ttl: TTL for new records. (optional)
        :type  ttl: ``int``

        :param extra: Extra attributes ('description', 'expire_sec', \
        'master_ips','refresh_sec', 'retry_sec', 'soa_email','status', 'tags')

        :type  extra: ``dict``

        :rtype: :class:`Zone`
        """
        if not isinstance(zone, Zone):
            raise LinodeExceptionV4("Invalid zone instance")

        attr = {
            "domain": domain,
            "type": type,
        }

        if ttl is not None:
            attr["ttl_sec"] = ttl

        merge_valid_keys(params=attr, valid_keys=VALID_ZONE_EXTRA_PARAMS_V4, extra=extra)

        data = self.connection.request(
            "/v4/domains/%s" % zone.id, data=json.dumps(attr), method="PUT"
        ).object

        return self._to_zone(data)

    def update_record(self, record, name=None, type=None, data=None, extra=None):
        """
        Update an existing record.

        :param record: Record to update.
        :type  record: :class:`Record`

        :param name: Record name.This field's actual usage\
         depends on the type of record this represents.
        :type  name: ``str``

        :param type: DNS record type
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: Extra attributes.('port', 'priority', \
        'protocol', 'service', 'tag', 'ttl_sec', 'target', 'weight')
        :type  extra: ``dict``

        :rtype: :class:`Record`
        """
        if not isinstance(record, Record):
            raise LinodeExceptionV4("Invalid zone instance")

        zone = record.zone
        attr = {}

        if name is not None:
            attr["name"] = name

        if type is not None:
            attr["type"] = self.RECORD_TYPE_MAP[type]

        if data is not None:
            attr["target"] = data

        merge_valid_keys(params=attr, valid_keys=VALID_RECORD_EXTRA_PARAMS_V4, extra=extra)

        data = self.connection.request(
            "/v4/domains/{}/records/{}".format(zone.id, record.id),
            data=json.dumps(attr),
            method="PUT",
        ).object

        return self._to_record(data, zone=zone)

    def delete_zone(self, zone):
        """
        Delete a zone.

        This will delete all the records belonging to this zone.

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``
        """
        if not isinstance(zone, Zone):
            raise LinodeExceptionV4("Invalid zone instance")

        response = self.connection.request("/v4/domains/%s" % zone.id, method="DELETE")
        return response.status == httplib.OK

    def delete_record(self, record):
        """
        Delete a record.

        :param record: Record to delete.
        :type  record: :class:`Record`

        :rtype: ``bool``
        """
        if not isinstance(record, Record):
            raise LinodeExceptionV4("Invalid record instance")

        zone = record.zone

        response = self.connection.request(
            "/v4/domains/{}/records/{}".format(zone.id, record.id), method="DELETE"
        )
        return response.status == httplib.OK

    def _to_record(self, item, zone=None):
        extra = {
            "port": item["port"],
            "weight": item["weight"],
            "priority": item["priority"],
            "service": item["service"],
            "protocol": item["protocol"],
            "created": self._to_datetime(item["created"]),
            "updated": self._to_datetime(item["updated"]),
        }
        type = self._string_to_record_type(item["type"])
        record = Record(
            id=item["id"],
            name=item["name"],
            type=type,
            data=item["target"],
            zone=zone,
            driver=self,
            ttl=item["ttl_sec"],
            extra=extra,
        )
        return record

    def _to_zone(self, item):
        """
        Build an Zone object from the item dictionary.
        """
        extra = {
            "soa_email": item["soa_email"],
            "status": item["status"],
            "description": item["description"],
            "tags": item["tags"],
            "retry_sec": item["retry_sec"],
            "master_ips": item["master_ips"],
            "axfr_ips": item["axfr_ips"],
            "expire_sec": item["expire_sec"],
            "refresh_sec": item["refresh_sec"],
            "created": self._to_datetime(item["created"]),
            "updated": self._to_datetime(item["updated"]),
        }
        zone = Zone(
            id=item["id"],
            domain=item["domain"],
            type=item["type"],
            ttl=item["ttl_sec"],
            driver=self,
            extra=extra,
        )
        return zone

    def _to_datetime(self, strtime):
        return datetime.strptime(strtime, "%Y-%m-%dT%H:%M:%S")

    def _paginated_request(self, url, obj, params=None):
        """
        Perform multiple calls in order to have a full list of elements when
        the API responses are paginated.

        :param url: API endpoint
        :type url: ``str``

        :param obj: Result object key
        :type obj: ``str``

        :param params: Request parameters
        :type params: ``dict``

        :return: ``list`` of API response objects
        :rtype: ``list``
        """
        objects = []
        params = params if params is not None else {}

        ret = self.connection.request(url, params=params).object

        data = list(ret.get(obj, []))
        current_page = int(ret.get("page", 1))
        num_of_pages = int(ret.get("pages", 1))
        objects.extend(data)
        for page in range(current_page + 1, num_of_pages + 1):
            # add param to request next page
            params["page"] = page
            ret = self.connection.request(url, params=params).object
            data = list(ret.get(obj, []))
            objects.extend(data)
        return objects



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import json
import tempfile
from unittest import mock

from libcloud.test import MockHttp, unittest
from libcloud.utils.py3 import b, httplib
from libcloud.utils.files import exhaust_iterator
from libcloud.test.file_fixtures import StorageFileFixtures
from libcloud.storage.drivers.backblaze_b2 import BackblazeB2StorageDriver


class MockAuthConn(mock.Mock):
    account_id = "abcdefgh"


class BackblazeB2StorageDriverTestCase(unittest.TestCase):
    driver_klass = BackblazeB2StorageDriver
    driver_args = ("a", "b")

    def setUp(self):
        self.driver_klass.connectionCls.authCls.conn_class = BackblazeB2MockHttp
        self.driver_klass.connectionCls.conn_class = BackblazeB2MockHttp

        BackblazeB2MockHttp.type = None
        self.driver = self.driver_klass(*self.driver_args)

    def test_list_containers(self):
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 3)
        self.assertEqual(containers[0].name, "test00001")
        self.assertEqual(containers[0].extra["id"], "481c37de2e1ab3bf5e150710")
        self.assertEqual(containers[0].extra["bucketType"], "allPrivate")

    def test_list_container_objects(self):
        container = self.driver.list_containers()[0]
        objects = self.driver.list_container_objects(container=container)
        self.assertEqual(len(objects), 4)
        self.assertEqual(objects[0].name, "2.txt")
        self.assertEqual(objects[0].size, 2)
        self.assertEqual(objects[0].extra["fileId"], "abcd")
        self.assertEqual(objects[0].extra["uploadTimestamp"], 1450545966000)

    def test_get_container(self):
        container = self.driver.get_container("test00001")
        self.assertEqual(container.name, "test00001")
        self.assertEqual(container.extra["id"], "481c37de2e1ab3bf5e150710")
        self.assertEqual(container.extra["bucketType"], "allPrivate")

    def test_get_object(self):
        obj = self.driver.get_object("test00001", "2.txt")
        self.assertEqual(obj.name, "2.txt")
        self.assertEqual(obj.size, 2)
        self.assertEqual(obj.extra["fileId"], "abcd")
        self.assertEqual(obj.extra["uploadTimestamp"], 1450545966000)

    def test_create_container(self):
        container = self.driver.create_container(container_name="test0005")
        self.assertEqual(container.name, "test0005")
        self.assertEqual(container.extra["id"], "681c87aebeaa530f5e250710")
        self.assertEqual(container.extra["bucketType"], "allPrivate")

    def test_delete_container(self):
        container = self.driver.list_containers()[0]
        result = self.driver.delete_container(container=container)
        self.assertTrue(result)

    def test_download_object(self):
        container = self.driver.list_containers()[0]
        obj = self.driver.list_container_objects(container=container)[0]
        _, destination_path = tempfile.mkstemp()
        result = self.driver.download_object(
            obj=obj, destination_path=destination_path, overwrite_existing=True
        )
        self.assertTrue(result)

    def test_download_object_as_stream(self):
        container = self.driver.list_containers()[0]
        obj = self.driver.list_container_objects(container=container)[0]

        stream = self.driver.download_object_as_stream(obj=obj, chunk_size=1024)
        self.assertTrue(hasattr(stream, "__iter__"))
        self.assertEqual(exhaust_iterator(stream), b("ab"))

    def test_upload_object(self):
        file_path = os.path.abspath(__file__)
        container = self.driver.list_containers()[0]
        obj = self.driver.upload_object(
            file_path=file_path, container=container, object_name="test0007.txt"
        )
        self.assertEqual(obj.name, "test0007.txt")
        self.assertEqual(obj.size, 24)
        self.assertEqual(obj.extra["fileId"], "abcde")

    def test_upload_object_via_stream(self):
        container = self.driver.list_containers()[0]
        file_path = os.path.abspath(__file__)

        with open(file_path, "rb") as fp:
            iterator = iter(fp)

            obj = self.driver.upload_object_via_stream(
                iterator=iterator, container=container, object_name="test0007.txt"
            )

        self.assertEqual(obj.name, "test0007.txt")
        self.assertEqual(obj.size, 24)
        self.assertEqual(obj.extra["fileId"], "abcde")

    def test_upload_object_with_metadata(self):
        file_path = os.path.abspath(__file__)
        container = self.driver.list_containers()[0]
        obj = self.driver.upload_object(
            file_path=file_path,
            container=container,
            object_name="test0007.txt",
            extra={"meta_data": {"foo": "bar", "baz": 1}},
        )
        self.assertEqual(obj.name, "test0007.txt")
        self.assertEqual(obj.size, 24)
        self.assertEqual(obj.extra["fileId"], "abcde")

    def test_delete_object(self):
        container = self.driver.list_containers()[0]
        obj = self.driver.list_container_objects(container=container)[0]
        result = self.driver.delete_object(obj=obj)
        self.assertTrue(result)

    def test_ex_hide_object(self):
        container = self.driver.list_containers()[0]
        container_id = container.extra["id"]
        obj = self.driver.ex_hide_object(container_id=container_id, object_name="2.txt")
        self.assertEqual(obj.name, "2.txt")

    def test_ex_list_object_versions(self):
        container = self.driver.list_containers()[0]
        container_id = container.extra["id"]
        objects = self.driver.ex_list_object_versions(container_id=container_id)
        self.assertEqual(len(objects), 9)

    def test_ex_get_upload_data(self):
        container = self.driver.list_containers()[0]
        container_id = container.extra["id"]
        data = self.driver.ex_get_upload_data(container_id=container_id)
        self.assertEqual(data["authorizationToken"], "nope")
        self.assertEqual(data["bucketId"], "481c37de2e1ab3bf5e150710")
        self.assertEqual(
            data["uploadUrl"],
            "https://podxxx.backblaze.com/b2api/v1/b2_upload_file/abcd/defg",
        )

    def test_ex_get_upload_url(self):
        container = self.driver.list_containers()[0]
        container_id = container.extra["id"]
        url = self.driver.ex_get_upload_url(container_id=container_id)
        self.assertEqual(url, "https://podxxx.backblaze.com/b2api/v1/b2_upload_file/abcd/defg")


class BackblazeB2MockHttp(MockHttp):
    fixtures = StorageFileFixtures("backblaze_b2")

    def _b2api_v1_b2_authorize_account(self, method, url, body, headers):
        if method == "GET":
            body = json.dumps(
                {
                    "accountId": "test",
                    "apiUrl": "https://apiNNN.backblazeb2.com",
                    "downloadUrl": "https://f002.backblazeb2.com",
                    "authorizationToken": "test",
                }
            )
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _b2api_v1_b2_list_buckets(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("b2_list_buckets.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _b2api_v1_b2_list_file_names(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("b2_list_file_names.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _b2api_v1_b2_create_bucket(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("b2_create_bucket.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _b2api_v1_b2_delete_bucket(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("b2_delete_bucket.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _b2api_v1_b2_delete_file_version(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("b2_delete_file_version.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _b2api_v1_b2_get_upload_url(self, method, url, body, headers):
        # test_upload_object
        if method == "GET":
            body = self.fixtures.load("b2_get_upload_url.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _b2api_v1_b2_upload_file_abcd_defg(self, method, url, body, headers):
        # test_upload_object
        if method == "POST":
            body = self.fixtures.load("b2_upload_file.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _b2api_v1_b2_list_file_versions(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("b2_list_file_versions.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _b2api_v1_b2_hide_file(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("b2_hide_file.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _file_test00001_2_txt(self, method, url, body, headers):
        # test_download_object
        if method == "GET":
            body = "ab"
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.dns.types import RecordType
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DNS_PARAMS_LINODE
from libcloud.dns.drivers.linode import LinodeDNSDriver, LinodeDNSDriverV4
from libcloud.test.file_fixtures import DNSFileFixtures


class LinodeTests(unittest.TestCase):
    def setUp(self):
        LinodeDNSDriverV4.connectionCls.conn_class = LinodeMockHttpV4
        LinodeMockHttpV4.type = None
        self.driver = LinodeDNSDriver(*DNS_PARAMS_LINODE)

    def test_correct_class_is_used(self):
        self.assertIsInstance(self.driver, LinodeDNSDriverV4)

    def test_unknown_api_version(self):
        self.assertRaises(NotImplementedError, LinodeDNSDriver, "foo", api_version="2.0")

    def test_list_zones(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 3)
        zone = zones[0]
        self.assertEqual(zone.id, "123")
        self.assertEqual(zone.domain, "test.com")
        self.assertEqual(zone.ttl, 300)
        self.assertEqual(zone.extra["soa_email"], "admin@test.com")

    def test_list_records(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone)
        self.assertEqual(len(records), 6)
        record = records[0]
        self.assertEqual(record.id, "123")
        self.assertEqual(record.data, "mail.test.com")
        self.assertEqual(record.ttl, 300)

    def test_get_zone(self):
        zone = self.driver.get_zone("123")
        self.assertEqual(zone.id, "123")
        self.assertEqual(zone.domain, "test.com")
        self.assertEqual(zone.extra["soa_email"], "admin@test.com")

    def test_get_zone_not_found(self):
        LinodeMockHttpV4.type = "ZONE_DOES_NOT_EXIST"
        with self.assertRaises(Exception):
            self.driver.get_zone("123")

    def test_get_record_A_RECORD(self):
        LinodeMockHttpV4.type = "A_RECORD"
        record = self.driver.get_record("123", "123")
        self.assertEqual(record.id, "123")
        self.assertEqual(record.name, "test.example.com")
        self.assertEqual(record.type, "A")

    def test_get_record_MX_RECORD(self):
        LinodeMockHttpV4.type = "MX_RECORD"
        record = self.driver.get_record("123", "123")
        self.assertEqual(record.id, "123")
        self.assertEqual(record.data, "mail.example.com")
        self.assertEqual(record.type, "MX")

    def test_create_zone(self):
        domain = "example.com"
        ttl = 300
        extra = {"soa_email": "admin@example.com"}
        zone = self.driver.create_zone(domain=domain, ttl=ttl, extra=extra)
        self.assertEqual(zone.ttl, 300)
        self.assertEqual(zone.domain, "example.com")
        self.assertEqual(zone.extra["soa_email"], "admin@example.com")

    def test_create_record(self):
        zone = self.driver.list_zones()[0]
        name = "test"
        type = RecordType.A
        data = "200.150.100.50"
        record = self.driver.create_record(name, zone, type, data)
        self.assertEqual(record.id, "123")
        self.assertEqual(record.name, name)
        self.assertEqual(record.type, "A")
        self.assertEqual(record.data, data)

    def test_update_zone(self):
        zone = self.driver.list_zones()[0]
        domain = "example.com"
        ttl = 300
        extra = {"description": "Testing", "soa_email": "admin@example.com"}
        updated_zone = self.driver.update_zone(zone, domain, ttl=ttl, extra=extra)
        self.assertEqual(updated_zone.domain, domain)
        self.assertEqual(updated_zone.ttl, ttl)
        self.assertEqual(updated_zone.extra["soa_email"], extra["soa_email"])
        self.assertEqual(updated_zone.extra["description"], extra["description"])

    def test_update_record(self):
        LinodeMockHttpV4.type = "A_RECORD"
        record = self.driver.get_record("123", "123")
        name = "test"
        data = "200.150.100.50"
        extra = {"ttl_sec": 3600}
        updated = self.driver.update_record(record, name=name, data=data, extra=extra)
        self.assertEqual(updated.name, name)
        self.assertEqual(updated.ttl, extra["ttl_sec"])
        self.assertEqual(updated.data, data)

    def test_delete_zone(self):
        zone = self.driver.list_zones()[0]
        self.assertTrue(self.driver.delete_zone(zone))

    def test_delete_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone)[0]
        self.assertTrue(self.driver.delete_record(record))


class LinodeMockHttpV4(MockHttp):
    fixtures = DNSFileFixtures("linode_v4")

    def _v4_domains(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("list_zones.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        if method == "POST":
            body = self.fixtures.load("create_zone.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v4_domains_123_records(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("list_records.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        if method == "POST":
            body = self.fixtures.load("create_record.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v4_domains_123(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("get_zone.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        if method == "PUT":
            body = self.fixtures.load("update_zone.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        if method == "DELETE":
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v4_domains_123_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = '{ "errors":[{"reason":"Not found"}]}'
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _v4_domains_123_A_RECORD(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v4_domains_123_MX_RECORD(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v4_domains_123_records_123_A_RECORD(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("get_record_A.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        if method == "PUT":
            body = self.fixtures.load("update_record.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v4_domains_123_records_123_MX_RECORD(self, method, url, body, headers):
        body = self.fixtures.load("get_record_MX.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v4_domains_123_records_123(self, method, url, body, headers):
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and


import sys
import datetime
import tempfile
from unittest.mock import Mock

from libcloud import __version__
from libcloud.test import unittest
from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import RecordType
from libcloud.utils.py3 import assertRegex

MOCK_RECORDS_VALUES = [
    {"id": 1, "name": "www", "type": RecordType.A, "data": "127.0.0.1"},
    {"id": 2, "name": "www", "type": RecordType.AAAA, "data": "2a01:4f8:121:3121::2"},
    # Custom TTL
    {
        "id": 3,
        "name": "www",
        "type": RecordType.A,
        "data": "127.0.0.1",
        "extra": {"ttl": 123},
    },
    # Record without a name
    {"id": 4, "name": "", "type": RecordType.A, "data": "127.0.0.1"},
    {"id": 5, "name": "test1", "type": RecordType.TXT, "data": "test foo bar"},
    # TXT record with quotes
    {"id": 5, "name": "test2", "type": RecordType.TXT, "data": 'test "foo" "bar"'},
    # Records with priority
    {
        "id": 5,
        "name": "",
        "type": RecordType.MX,
        "data": "mx.example.com",
        "extra": {"priority": 10},
    },
    {
        "id": 5,
        "name": "",
        "type": RecordType.SRV,
        "data": "10 3333 example.com",
        "extra": {"priority": 20},
    },
]


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        self.driver = DNSDriver("none", "none")
        self.tmp_file = tempfile.mkstemp()
        self.tmp_path = self.tmp_file[1]

    def test_export_zone_to_bind_format_slave_should_throw(self):
        zone = Zone(id=1, domain="example.com", type="slave", ttl=900, driver=self.driver)
        self.assertRaises(ValueError, zone.export_to_bind_format)

    def test_export_zone_to_bind_format_success(self):
        zone = Zone(id=1, domain="example.com", type="master", ttl=900, driver=self.driver)

        mock_records = []

        for values in MOCK_RECORDS_VALUES:
            values = values.copy()
            values["driver"] = self.driver
            values["zone"] = zone
            record = Record(**values)
            mock_records.append(record)

        self.driver.list_records = Mock()
        self.driver.list_records.return_value = mock_records

        now = datetime.datetime.utcnow()

        result = self.driver.export_zone_to_bind_format(zone=zone)
        self.driver.export_zone_to_bind_zone_file(zone=zone, file_path=self.tmp_path)

        with open(self.tmp_path) as fp:
            content = fp.read()

        lines1 = result.split("\n")
        lines2 = content.split("\n")

        date_str = "{}-{}-{} {}:{}:{}".format(
            now.year,
            zero_pad(now.month),
            zero_pad(now.day),
            zero_pad(now.hour),
            zero_pad(now.minute),
            zero_pad(now.second),
        )
        expected_header = "; Generated by Libcloud v{} on {} UTC".format(
            __version__,
            date_str,
        )

        self.assertEqual(lines1[0], expected_header)
        self.assertEqual(lines2[0], expected_header)

        for lines in [lines1, lines2]:
            self.assertEqual(len(lines), 2 + 1 + 9)
            assertRegex(self, lines[1], r"\$ORIGIN example\.com\.")
            assertRegex(self, lines[2], r"\$TTL 900")

            assertRegex(self, lines[4], r"www.example.com\.\s+900\s+IN\s+A\s+127\.0\.0\.1")
            assertRegex(
                self,
                lines[5],
                r"www.example.com\.\s+900\s+IN\s+AAAA\s+2a01:4f8:121:3121::2",
            )
            assertRegex(self, lines[6], r"www.example.com\.\s+123\s+IN\s+A\s+127\.0\.0\.1")
            assertRegex(self, lines[7], r"example.com\.\s+900\s+IN\s+A\s+127\.0\.0\.1")
            assertRegex(self, lines[8], r'test1.example.com\.\s+900\s+IN\s+TXT\s+"test foo bar"')
            assertRegex(
                self,
                lines[9],
                r'test2.example.com\.\s+900\s+IN\s+TXT\s+"test \\"foo\\" \\"bar\\""',
            )
            assertRegex(self, lines[10], r"example.com\.\s+900\s+IN\s+MX\s+10\s+mx.example.com")
            assertRegex(
                self,
                lines[11],
                r"example.com\.\s+900\s+IN\s+SRV\s+20\s+10 3333 example.com",
            )

    def test_get_numeric_id(self):
        values = MOCK_RECORDS_VALUES[0].copy()
        values["driver"] = self.driver
        values["zone"] = None
        record = Record(**values)

        record.id = "abcd"
        result = record._get_numeric_id()
        self.assertEqual(result, "abcd")

        record.id = "1"
        result = record._get_numeric_id()
        self.assertEqual(result, 1)

        record.id = "12345"
        result = record._get_numeric_id()
        self.assertEqual(result, 12345)

        record.id = ""
        result = record._get_numeric_id()
        self.assertEqual(result, "")

        record.id = None
        result = record._get_numeric_id()
        self.assertEqual(result, "")


def zero_pad(value: int) -> str:
    if value < 10:
        return "0" + str(value)

    return str(value)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import hmac
import atexit
from time import time
from hashlib import sha1

from libcloud.utils.py3 import b, httplib, urlquote, urlencode
from libcloud.common.base import Response, RawResponse
from libcloud.utils.files import read_in_chunks
from libcloud.common.types import LibcloudError, MalformedResponseError
from libcloud.storage.base import Object, Container, StorageDriver
from libcloud.storage.types import (
    ObjectDoesNotExistError,
    ObjectHashMismatchError,
    ContainerIsNotEmptyError,
    InvalidContainerNameError,
    ContainerDoesNotExistError,
    ContainerAlreadyExistsError,
)
from libcloud.common.openstack import OpenStackDriverMixin, OpenStackBaseConnection
from libcloud.common.rackspace import AUTH_URL
from libcloud.storage.providers import Provider

try:
    import simplejson as json
except ImportError:
    import json  # type: ignore

from io import FileIO as file

CDN_HOST = "cdn.clouddrive.com"
API_VERSION = "v1.0"

# Keys which are used to select a correct endpoint from the service catalog.
INTERNAL_ENDPOINT_KEY = "internalURL"
PUBLIC_ENDPOINT_KEY = "publicURL"


class CloudFilesResponse(Response):
    valid_response_codes = [httplib.NOT_FOUND, httplib.CONFLICT]

    def success(self):
        i = int(self.status)
        return 200 <= i <= 299 or i in self.valid_response_codes

    def parse_body(self):
        if not self.body:
            return None

        if "content-type" in self.headers:
            key = "content-type"
        elif "Content-Type" in self.headers:
            key = "Content-Type"
        else:
            raise LibcloudError("Missing content-type header")

        content_type = self.headers[key]
        if content_type.find(";") != -1:
            content_type = content_type.split(";")[0]

        if content_type == "application/json":
            try:
                data = json.loads(self.body)
            except Exception:
                raise MalformedResponseError(
                    "Failed to parse JSON",
                    body=self.body,
                    driver=CloudFilesStorageDriver,
                )
        elif content_type == "text/plain":
            data = self.body
        else:
            data = self.body

        return data


class CloudFilesRawResponse(CloudFilesResponse, RawResponse):
    pass


class OpenStackSwiftConnection(OpenStackBaseConnection):
    """
    Connection class for the OpenStack Swift endpoint.
    """

    responseCls = CloudFilesResponse
    rawResponseCls = CloudFilesRawResponse

    auth_url = AUTH_URL
    _auth_version = "1.0"

    # TODO: Reverse the relationship - Swift -> CloudFiles
    def __init__(self, user_id, key, secure=True, **kwargs):
        # Ignore this for now
        kwargs.pop("use_internal_url", None)
        super().__init__(user_id, key, secure=secure, **kwargs)
        self.api_version = API_VERSION
        self.accept_format = "application/json"

        self._service_type = self._ex_force_service_type or "object-store"
        self._service_name = self._ex_force_service_name or "swift"

        if self._ex_force_service_region:
            self._service_region = self._ex_force_service_region
        else:
            self._service_region = None

    def get_endpoint(self, *args, **kwargs):
        if ("2.0" in self._auth_version) or ("3.x" in self._auth_version):
            endpoint = self.service_catalog.get_endpoint(
                service_type=self._service_type,
                name=self._service_name,
                region=self._service_region,
            )
        elif ("1.1" in self._auth_version) or ("1.0" in self._auth_version):
            endpoint = self.service_catalog.get_endpoint(
                name=self._service_name, region=self._service_region
            )
        else:
            endpoint = None

        if endpoint:
            return endpoint.url
        else:
            raise LibcloudError("Could not find specified endpoint")

    def request(
        self,
        action,
        params=None,
        data="",
        headers=None,
        method="GET",
        raw=False,
        cdn_request=False,
    ):
        if not headers:
            headers = {}
        if not params:
            params = {}

        self.cdn_request = cdn_request
        params["format"] = "json"

        if method in ["POST", "PUT"] and "Content-Type" not in headers:
            headers.update({"Content-Type": "application/json; charset=UTF-8"})

        return super().request(
            action=action,
            params=params,
            data=data,
            method=method,
            headers=headers,
            raw=raw,
        )


class CloudFilesConnection(OpenStackSwiftConnection):
    """
    Base connection class for the Cloudfiles driver.
    """

    responseCls = CloudFilesResponse
    rawResponseCls = CloudFilesRawResponse

    auth_url = AUTH_URL
    _auth_version = "2.0"

    def __init__(self, user_id, key, secure=True, use_internal_url=False, **kwargs):
        super().__init__(user_id, key, secure=secure, **kwargs)
        self.api_version = API_VERSION
        self.accept_format = "application/json"
        self.cdn_request = False
        self.use_internal_url = use_internal_url

    def get_endpoint(self):
        region = self._ex_force_service_region.upper()

        if "2.0" in self._auth_version:
            ep = self.service_catalog.get_endpoint(
                service_type="object-store",
                name="cloudFiles",
                region=region,
                endpoint_type="internal" if self.use_internal_url else "external",
            )
            cdn_ep = self.service_catalog.get_endpoint(
                service_type="rax:object-cdn",
                name="cloudFilesCDN",
                region=region,
                # Cloud Files has no concept of an "internal" CDN.
                endpoint_type="external",
            )
        else:
            raise LibcloudError('Auth version "%s" not supported' % (self._auth_version))

        # if this is a CDN request, return the cdn url instead
        if self.cdn_request:
            ep = cdn_ep

        if not ep or not ep.url:
            raise LibcloudError("Could not find specified endpoint")

        return ep.url

    def request(
        self,
        action,
        params=None,
        data="",
        headers=None,
        method="GET",
        raw=False,
        cdn_request=False,
    ):
        if not headers:
            headers = {}
        if not params:
            params = {}

        self.cdn_request = cdn_request
        params["format"] = "json"

        if method in ["POST", "PUT"] and "Content-Type" not in headers:
            headers.update({"Content-Type": "application/json; charset=UTF-8"})

        return super().request(
            action=action,
            params=params,
            data=data,
            method=method,
            headers=headers,
            raw=raw,
            cdn_request=cdn_request,
        )


class CloudFilesStorageDriver(StorageDriver, OpenStackDriverMixin):
    """
    CloudFiles driver.
    """

    name = "CloudFiles"
    website = "http://www.rackspace.com/"

    connectionCls = CloudFilesConnection
    hash_type = "md5"
    supports_chunked_encoding = True

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        region="ord",
        use_internal_url=False,
        **kwargs,
    ):
        """
        @inherits:  :class:`StorageDriver.__init__`

        :param region: ID of the region which should be used.
        :type region: ``str``
        """
        # This is here for backard compatibility
        if "ex_force_service_region" in kwargs:
            region = kwargs["ex_force_service_region"]

        self.use_internal_url = use_internal_url
        OpenStackDriverMixin.__init__(self, **kwargs)
        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            region=region,
            **kwargs,
        )

    @classmethod
    def list_regions(cls):
        return ["ord", "dfw", "iad", "lon", "hkg", "syd"]

    def iterate_containers(self):
        response = self.connection.request("")

        if response.status == httplib.NO_CONTENT:
            return []
        elif response.status == httplib.OK:
            return self._to_container_list(json.loads(response.body))

        raise LibcloudError("Unexpected status code: %s" % (response.status))

    def get_container(self, container_name):
        container_name_encoded = self._encode_container_name(container_name)
        response = self.connection.request("/%s" % (container_name_encoded), method="HEAD")

        if response.status == httplib.NO_CONTENT:
            container = self._headers_to_container(container_name, response.headers)
            return container
        elif response.status == httplib.NOT_FOUND:
            raise ContainerDoesNotExistError(None, self, container_name)

        raise LibcloudError("Unexpected status code: %s" % (response.status))

    def get_object(self, container_name, object_name):
        container = self.get_container(container_name)
        container_name_encoded = self._encode_container_name(container_name)
        object_name_encoded = self._encode_object_name(object_name)

        response = self.connection.request(
            "/{}/{}".format(container_name_encoded, object_name_encoded), method="HEAD"
        )
        if response.status in [httplib.OK, httplib.NO_CONTENT]:
            obj = self._headers_to_object(object_name, container, response.headers)
            return obj
        elif response.status == httplib.NOT_FOUND:
            raise ObjectDoesNotExistError(None, self, object_name)

        raise LibcloudError("Unexpected status code: %s" % (response.status))

    def get_container_cdn_url(self, container, ex_ssl_uri=False):
        # pylint: disable=unexpected-keyword-arg
        container_name_encoded = self._encode_container_name(container.name)
        response = self.connection.request(
            "/%s" % (container_name_encoded), method="HEAD", cdn_request=True
        )

        if response.status == httplib.NO_CONTENT:
            if ex_ssl_uri:
                cdn_url = response.headers["x-cdn-ssl-uri"]
            else:
                cdn_url = response.headers["x-cdn-uri"]
            return cdn_url
        elif response.status == httplib.NOT_FOUND:
            raise ContainerDoesNotExistError(value="", container_name=container.name, driver=self)

        raise LibcloudError("Unexpected status code: %s" % (response.status))

    def get_object_cdn_url(self, obj):
        container_cdn_url = self.get_container_cdn_url(container=obj.container)
        return "{}/{}".format(container_cdn_url, obj.name)

    def enable_container_cdn(self, container, ex_ttl=None):
        """
        @inherits: :class:`StorageDriver.enable_container_cdn`

        :param ex_ttl: cache time to live
        :type ex_ttl: ``int``
        """
        container_name = self._encode_container_name(container.name)
        headers = {"X-CDN-Enabled": "True"}

        if ex_ttl:
            headers["X-TTL"] = ex_ttl

        # pylint: disable=unexpected-keyword-arg
        response = self.connection.request(
            "/%s" % (container_name), method="PUT", headers=headers, cdn_request=True
        )

        return response.status in [httplib.CREATED, httplib.ACCEPTED]

    def create_container(self, container_name):
        container_name_encoded = self._encode_container_name(container_name)
        response = self.connection.request("/%s" % (container_name_encoded), method="PUT")

        if response.status == httplib.CREATED:
            # Accepted mean that container is not yet created but it will be
            # eventually
            extra = {"object_count": 0}
            container = Container(name=container_name, extra=extra, driver=self)

            return container
        elif response.status == httplib.ACCEPTED:
            error = ContainerAlreadyExistsError(None, self, container_name)
            raise error

        raise LibcloudError("Unexpected status code: %s" % (response.status))

    def delete_container(self, container):
        name = self._encode_container_name(container.name)

        # Only empty container can be deleted
        response = self.connection.request("/%s" % (name), method="DELETE")

        if response.status == httplib.NO_CONTENT:
            return True
        elif response.status == httplib.NOT_FOUND:
            raise ContainerDoesNotExistError(value="", container_name=name, driver=self)
        elif response.status == httplib.CONFLICT:
            # @TODO: Add "delete_all_objects" parameter?
            raise ContainerIsNotEmptyError(value="", container_name=name, driver=self)

    def download_object(
        self, obj, destination_path, overwrite_existing=False, delete_on_failure=True
    ):
        container_name = obj.container.name
        object_name = obj.name
        response = self.connection.request(
            "/{}/{}".format(container_name, object_name), method="GET", raw=True
        )

        return self._get_object(
            obj=obj,
            callback=self._save_object,
            response=response,
            callback_kwargs={
                "obj": obj,
                "response": response.response,
                "destination_path": destination_path,
                "overwrite_existing": overwrite_existing,
                "delete_on_failure": delete_on_failure,
            },
            success_status_code=httplib.OK,
        )

    def download_object_as_stream(self, obj, chunk_size=None):
        container_name = obj.container.name
        object_name = obj.name
        response = self.connection.request(
            "/{}/{}".format(container_name, object_name), method="GET", raw=True
        )

        return self._get_object(
            obj=obj,
            callback=read_in_chunks,
            response=response,
            callback_kwargs={
                "iterator": response.iter_content(chunk_size),
                "chunk_size": chunk_size,
            },
            success_status_code=httplib.OK,
        )

    def download_object_range(
        self,
        obj,
        destination_path,
        start_bytes,
        end_bytes=None,
        overwrite_existing=False,
        delete_on_failure=True,
    ):
        self._validate_start_and_end_bytes(start_bytes=start_bytes, end_bytes=end_bytes)

        container_name = obj.container.name
        object_name = obj.name
        headers = {"Range": self._get_standard_range_str(start_bytes, end_bytes)}
        response = self.connection.request(
            "/{}/{}".format(container_name, object_name),
            method="GET",
            headers=headers,
            raw=True,
        )

        return self._get_object(
            obj=obj,
            callback=self._save_object,
            response=response,
            callback_kwargs={
                "obj": obj,
                "response": response.response,
                "destination_path": destination_path,
                "overwrite_existing": overwrite_existing,
                "delete_on_failure": delete_on_failure,
                "partial_download": True,
            },
            success_status_code=httplib.PARTIAL_CONTENT,
        )

    def download_object_range_as_stream(self, obj, start_bytes, end_bytes=None, chunk_size=None):
        self._validate_start_and_end_bytes(start_bytes=start_bytes, end_bytes=end_bytes)
        container_name = obj.container.name
        object_name = obj.name
        headers = {"Range": self._get_standard_range_str(start_bytes, end_bytes)}
        response = self.connection.request(
            "/{}/{}".format(container_name, object_name),
            headers=headers,
            method="GET",
            raw=True,
        )

        return self._get_object(
            obj=obj,
            callback=read_in_chunks,
            response=response,
            callback_kwargs={
                "iterator": response.iter_content(chunk_size),
                "chunk_size": chunk_size,
            },
            success_status_code=httplib.PARTIAL_CONTENT,
        )

    def upload_object(
        self,
        file_path,
        container,
        object_name,
        extra=None,
        verify_hash=True,
        headers=None,
    ):
        """
        Upload an object.

        Note: This will override file with a same name if it already exists.
        """

        return self._put_object(
            container=container,
            object_name=object_name,
            extra=extra,
            file_path=file_path,
            verify_hash=verify_hash,
            headers=headers,
        )

    def upload_object_via_stream(self, iterator, container, object_name, extra=None, headers=None):
        if isinstance(iterator, file):
            iterator = iter(iterator)

        return self._put_object(
            container=container,
            object_name=object_name,
            extra=extra,
            stream=iterator,
            headers=headers,
        )

    def delete_object(self, obj):
        container_name = self._encode_container_name(obj.container.name)
        object_name = self._encode_object_name(obj.name)

        response = self.connection.request(
            "/{}/{}".format(container_name, object_name), method="DELETE"
        )

        if response.status == httplib.NO_CONTENT:
            return True
        elif response.status == httplib.NOT_FOUND:
            raise ObjectDoesNotExistError(value="", object_name=object_name, driver=self)

        raise LibcloudError("Unexpected status code: %s" % (response.status))

    def ex_purge_object_from_cdn(self, obj, email=None):
        """
        Purge edge cache for the specified object.

        :param email: Email where a notification will be sent when the job
        completes. (optional)
        :type email: ``str``
        """
        container_name = self._encode_container_name(obj.container.name)
        object_name = self._encode_object_name(obj.name)
        headers = {"X-Purge-Email": email} if email else {}

        # pylint: disable=unexpected-keyword-arg
        response = self.connection.request(
            "/{}/{}".format(container_name, object_name),
            method="DELETE",
            headers=headers,
            cdn_request=True,
        )

        return response.status == httplib.NO_CONTENT

    def ex_get_meta_data(self):
        """
        Get meta data

        :rtype: ``dict``
        """
        response = self.connection.request("", method="HEAD")

        if response.status == httplib.NO_CONTENT:
            container_count = response.headers.get("x-account-container-count", "unknown")
            object_count = response.headers.get("x-account-object-count", "unknown")
            bytes_used = response.headers.get("x-account-bytes-used", "unknown")
            temp_url_key = response.headers.get("x-account-meta-temp-url-key", None)

            return {
                "container_count": int(container_count),
                "object_count": int(object_count),
                "bytes_used": int(bytes_used),
                "temp_url_key": temp_url_key,
            }

        raise LibcloudError("Unexpected status code: %s" % (response.status))

    def ex_multipart_upload_object(
        self,
        file_path,
        container,
        object_name,
        chunk_size=33554432,
        extra=None,
        verify_hash=True,
    ):
        object_size = os.path.getsize(file_path)
        if object_size < chunk_size:
            return self.upload_object(
                file_path, container, object_name, extra=extra, verify_hash=verify_hash
            )

        iter_chunk_reader = FileChunkReader(file_path, chunk_size)

        for index, iterator in enumerate(iter_chunk_reader):
            self._upload_object_part(
                container=container,
                object_name=object_name,
                part_number=index,
                iterator=iterator,
                verify_hash=verify_hash,
            )

        return self._upload_object_manifest(
            container=container,
            object_name=object_name,
            extra=extra,
            verify_hash=verify_hash,
        )

    def ex_enable_static_website(self, container, index_file="index.html"):
        """
        Enable serving a static website.

        :param container: Container instance
        :type container: :class:`Container`

        :param index_file: Name of the object which becomes an index page for
        every sub-directory in this container.
        :type index_file: ``str``

        :rtype: ``bool``
        """
        container_name = container.name
        headers = {"X-Container-Meta-Web-Index": index_file}

        # pylint: disable=unexpected-keyword-arg
        response = self.connection.request(
            "/%s" % (container_name), method="POST", headers=headers, cdn_request=False
        )

        return response.status in [httplib.CREATED, httplib.ACCEPTED]

    def ex_set_error_page(self, container, file_name="error.html"):
        """
        Set a custom error page which is displayed if file is not found and
        serving of a static website is enabled.

        :param container: Container instance
        :type container: :class:`Container`

        :param file_name: Name of the object which becomes the error page.
        :type file_name: ``str``

        :rtype: ``bool``
        """
        container_name = container.name
        headers = {"X-Container-Meta-Web-Error": file_name}

        # pylint: disable=unexpected-keyword-arg
        response = self.connection.request(
            "/%s" % (container_name), method="POST", headers=headers, cdn_request=False
        )

        return response.status in [httplib.CREATED, httplib.ACCEPTED]

    def ex_set_account_metadata_temp_url_key(self, key):
        """
        Set the metadata header X-Account-Meta-Temp-URL-Key on your Cloud
        Files account.

        :param key: X-Account-Meta-Temp-URL-Key
        :type key: ``str``

        :rtype: ``bool``
        """
        headers = {"X-Account-Meta-Temp-URL-Key": key}

        # pylint: disable=unexpected-keyword-arg
        response = self.connection.request("", method="POST", headers=headers, cdn_request=False)

        return response.status in [
            httplib.OK,
            httplib.NO_CONTENT,
            httplib.CREATED,
            httplib.ACCEPTED,
        ]

    def ex_get_object_temp_url(self, obj, method="GET", timeout=60):
        """
        Create a temporary URL to allow others to retrieve or put objects
        in your Cloud Files account for as long or as short a time as you
        wish.  This method is specifically for allowing users to retrieve
        or update an object.

        :param obj: The object that you wish to make temporarily public
        :type obj: :class:`Object`

        :param method: Which method you would like to allow, 'PUT' or 'GET'
        :type method: ``str``

        :param timeout: Time (in seconds) after which you want the TempURL
        to expire.
        :type timeout: ``int``

        :rtype: ``bool``
        """
        # pylint: disable=no-member
        self.connection._populate_hosts_and_request_paths()
        expires = int(time() + timeout)
        path = "{}/{}/{}".format(self.connection.request_path, obj.container.name, obj.name)
        try:
            key = self.ex_get_meta_data()["temp_url_key"]
            assert key is not None
        except Exception:
            raise KeyError(
                "You must first set the "
                + "X-Account-Meta-Temp-URL-Key header on your "
                + "Cloud Files account using "
                + "ex_set_account_metadata_temp_url_key before "
                + "you can use this method."
            )
        hmac_body = "{}\n{}\n{}".format(method, expires, path)
        sig = hmac.new(b(key), b(hmac_body), sha1).hexdigest()
        params = urlencode({"temp_url_sig": sig, "temp_url_expires": expires})

        temp_url = "https://{}/{}/{}?{}".format(
            self.connection.host + self.connection.request_path,
            obj.container.name,
            obj.name,
            params,
        )

        return temp_url

    def _upload_object_part(self, container, object_name, part_number, iterator, verify_hash=True):
        part_name = object_name + "/%08d" % part_number
        extra = {"content_type": "application/octet-stream"}

        self._put_object(
            container=container,
            object_name=part_name,
            extra=extra,
            stream=iterator,
            verify_hash=verify_hash,
        )

    def _upload_object_manifest(self, container, object_name, extra=None, verify_hash=True):
        extra = extra or {}
        meta_data = extra.get("meta_data")

        container_name_encoded = self._encode_container_name(container.name)
        object_name_encoded = self._encode_object_name(object_name)
        request_path = "/{}/{}".format(container_name_encoded, object_name_encoded)

        # pylint: disable=no-member
        headers = {
            "X-Auth-Token": self.connection.auth_token,
            "X-Object-Manifest": "{}/{}/".format(container_name_encoded, object_name_encoded),
        }

        data = ""
        response = self.connection.request(
            request_path, method="PUT", data=data, headers=headers, raw=True
        )

        object_hash = None

        if verify_hash:
            hash_function = self._get_hash_function()
            hash_function.update(b(data))
            data_hash = hash_function.hexdigest()
            object_hash = response.headers.get("etag")

            if object_hash != data_hash:
                raise ObjectHashMismatchError(
                    value=("MD5 hash checksum does not match (expected=%s, " + "actual=%s)")
                    % (data_hash, object_hash),
                    object_name=object_name,
                    driver=self,
                )

        obj = Object(
            name=object_name,
            size=0,
            hash=object_hash,
            extra=None,
            meta_data=meta_data,
            container=container,
            driver=self,
        )

        return obj

    def iterate_container_objects(self, container, prefix=None, ex_prefix=None):
        """
        Return a generator of objects for the given container.

        :param container: Container instance
        :type container: :class:`Container`

        :param prefix: Only get objects with names starting with prefix
        :type prefix: ``str``

        :param ex_prefix: (Deprecated.) Only get objects with names starting
                          with ex_prefix
        :type ex_prefix: ``str``

        :return: A generator of Object instances.
        :rtype: ``generator`` of :class:`Object`
        """
        prefix = self._normalize_prefix_argument(prefix, ex_prefix)

        params = {}

        if prefix:
            params["prefix"] = prefix

        while True:
            container_name_encoded = self._encode_container_name(container.name)
            response = self.connection.request("/%s" % (container_name_encoded), params=params)

            if response.status == httplib.NO_CONTENT:
                # Empty or non-existent container
                break
            elif response.status == httplib.OK:
                objects = self._to_object_list(json.loads(response.body), container)

                if len(objects) == 0:
                    break

                for obj in objects:
                    yield obj
                params["marker"] = obj.name

            else:
                raise LibcloudError("Unexpected status code: %s" % (response.status))

    def _put_object(
        self,
        container,
        object_name,
        extra=None,
        file_path=None,
        stream=None,
        verify_hash=True,
        headers=None,
    ):
        extra = extra or {}
        container_name_encoded = self._encode_container_name(container.name)
        object_name_encoded = self._encode_object_name(object_name)
        content_type = extra.get("content_type", None)
        meta_data = extra.get("meta_data", None)
        content_disposition = extra.get("content_disposition", None)

        headers = headers or {}
        if meta_data:
            for key, value in list(meta_data.items()):
                key = "X-Object-Meta-%s" % (key)
                headers[key] = value

        if content_disposition is not None:
            headers["Content-Disposition"] = content_disposition

        request_path = "/{}/{}".format(container_name_encoded, object_name_encoded)
        result_dict = self._upload_object(
            object_name=object_name,
            content_type=content_type,
            request_path=request_path,
            request_method="PUT",
            headers=headers,
            file_path=file_path,
            stream=stream,
        )

        response = result_dict["response"]
        bytes_transferred = result_dict["bytes_transferred"]
        server_hash = result_dict["response"].headers.get("etag", None)

        if response.status == httplib.EXPECTATION_FAILED:
            raise LibcloudError(value="Missing content-type header", driver=self)
        elif verify_hash and not server_hash:
            raise LibcloudError(value="Server didn't return etag", driver=self)
        elif verify_hash and result_dict["data_hash"] != server_hash:
            raise ObjectHashMismatchError(
                value=("MD5 hash checksum does not match (expected=%s, " + "actual=%s)")
                % (result_dict["data_hash"], server_hash),
                object_name=object_name,
                driver=self,
            )
        elif response.status == httplib.CREATED:
            obj = Object(
                name=object_name,
                size=bytes_transferred,
                hash=server_hash,
                extra=None,
                meta_data=meta_data,
                container=container,
                driver=self,
            )

            return obj
        else:
            # @TODO: Add test case for this condition (probably 411)
            raise LibcloudError("status_code=%s" % (response.status), driver=self)

    def _encode_container_name(self, name):
        """
        Encode container name so it can be used as part of the HTTP request.
        """
        if name.startswith("/"):
            name = name[1:]
        name = urlquote(name)

        if name.find("/") != -1:
            raise InvalidContainerNameError(
                value="Container name cannot" " contain slashes",
                container_name=name,
                driver=self,
            )

        if len(name) > 256:
            raise InvalidContainerNameError(
                value="Container name cannot be longer than 256 bytes",
                container_name=name,
                driver=self,
            )

        return name

    def _encode_object_name(self, name):
        name = urlquote(name)
        return name

    def _to_container_list(self, response):
        # @TODO: Handle more than 10k containers - use "lazy list"?
        for container in response:
            extra = {
                "object_count": int(container["count"]),
                "size": int(container["bytes"]),
            }
            yield Container(name=container["name"], extra=extra, driver=self)

    def _to_object_list(self, response, container):
        objects = []

        for obj in response:
            name = obj["name"]
            size = int(obj["bytes"])
            hash = obj["hash"]
            extra = {
                "content_type": obj["content_type"],
                "last_modified": obj["last_modified"],
            }
            objects.append(
                Object(
                    name=name,
                    size=size,
                    hash=hash,
                    extra=extra,
                    meta_data=None,
                    container=container,
                    driver=self,
                )
            )

        return objects

    def _headers_to_container(self, name, headers):
        size = int(headers.get("x-container-bytes-used", 0))
        object_count = int(headers.get("x-container-object-count", 0))

        extra = {"object_count": object_count, "size": size}
        container = Container(name=name, extra=extra, driver=self)
        return container

    def _headers_to_object(self, name, container, headers):
        size = int(headers.pop("content-length", 0))
        last_modified = headers.pop("last-modified", None)
        etag = headers.pop("etag", None)
        content_type = headers.pop("content-type", None)

        meta_data = {}
        for key, value in list(headers.items()):
            if key.find("x-object-meta-") != -1:
                key = key.replace("x-object-meta-", "")
                meta_data[key] = value

        extra = {"content_type": content_type, "last_modified": last_modified}

        obj = Object(
            name=name,
            size=size,
            hash=etag,
            extra=extra,
            meta_data=meta_data,
            container=container,
            driver=self,
        )
        return obj

    def _ex_connection_class_kwargs(self):
        kwargs = self.openstack_connection_kwargs()
        kwargs["ex_force_service_region"] = self.region
        kwargs["use_internal_url"] = self.use_internal_url
        return kwargs


class OpenStackSwiftStorageDriver(CloudFilesStorageDriver):
    """
    Storage driver for the OpenStack Swift.
    """

    type = Provider.CLOUDFILES_SWIFT
    name = "OpenStack Swift"
    connectionCls = OpenStackSwiftConnection

    # TODO: Reverse the relationship - Swift -> CloudFiles

    def __init__(self, key, secret=None, secure=True, host=None, port=None, region=None, **kwargs):
        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            region=region,
            **kwargs,
        )


class FileChunkReader:
    def __init__(self, file_path, chunk_size):
        self.file_path = file_path
        self.total = os.path.getsize(file_path)
        self.chunk_size = chunk_size
        self.bytes_read = 0
        self.stop_iteration = False

    def __iter__(self):
        return self

    def next(self):
        if self.stop_iteration:
            raise StopIteration

        start_block = self.bytes_read
        end_block = start_block + self.chunk_size
        if end_block >= self.total:
            end_block = self.total
            self.stop_iteration = True
        self.bytes_read += end_block - start_block
        return ChunkStreamReader(
            file_path=self.file_path,
            start_block=start_block,
            end_block=end_block,
            chunk_size=8192,
        )

    def __next__(self):
        return self.next()


class ChunkStreamReader:
    def __init__(self, file_path, start_block, end_block, chunk_size):
        self.fd = open(file_path, "rb")
        self.fd.seek(start_block)
        self.start_block = start_block
        self.end_block = end_block
        self.chunk_size = chunk_size
        self.bytes_read = 0
        self.stop_iteration = False

        # Work around to make sure file description is closed even if the
        # iterator is never read from or if it's not fully exhausted
        def close_file(fd):
            try:
                fd.close()
            except Exception:
                pass

        atexit.register(close_file, self.fd)

    def __iter__(self):
        return self

    def next(self):
        if self.stop_iteration:
            self.fd.close()
            raise StopIteration

        block_size = self.chunk_size
        if self.bytes_read + block_size > self.end_block - self.start_block:
            block_size = self.end_block - self.start_block - self.bytes_read
            self.stop_iteration = True

        block = self.fd.read(block_size)
        self.bytes_read += block_size
        return block

    def __next__(self):
        return self.next()



# Copyright 2002-2009, Distributed Systems Architecture Group, Universidad
# Complutense de Madrid (dsa-research.org)
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
OpenNebula.org driver.
"""

__docformat__ = "epytext"

import hashlib
from base64 import b64encode

from libcloud.utils.py3 import ET, b, next, httplib
from libcloud.common.base import XmlResponse, ConnectionUserAndKey
from libcloud.common.types import InvalidCredsError
from libcloud.compute.base import (
    Node,
    NodeSize,
    NodeImage,
    NodeState,
    NodeDriver,
    NodeLocation,
    StorageVolume,
)
from libcloud.compute.providers import Provider

__all__ = [
    "ACTION",
    "OpenNebulaResponse",
    "OpenNebulaConnection",
    "OpenNebulaNodeSize",
    "OpenNebulaNetwork",
    "OpenNebulaNodeDriver",
    "OpenNebula_1_4_NodeDriver",
    "OpenNebula_2_0_NodeDriver",
    "OpenNebula_3_0_NodeDriver",
    "OpenNebula_3_2_NodeDriver",
    "OpenNebula_3_8_NodeDriver",
]

API_HOST = ""
API_PORT = (4567, 443)
API_SECURE = True
API_PLAIN_AUTH = False
DEFAULT_API_VERSION = "3.2"


class ACTION:
    """
    All actions, except RESUME, only apply when the VM is in the "Running"
    state.
    """

    STOP = "STOPPED"
    """
    The VM is stopped, and its memory state stored to a checkpoint file. VM
    state, and disk image, are transferred back to the front-end. Resuming
    the VM requires the VM instance to be re-scheduled.
    """

    SUSPEND = "SUSPENDED"
    """
    The VM is stopped, and its memory state stored to a checkpoint file. The VM
    state, and disk image, are left on the host to be resumed later. Resuming
    the VM does not require the VM to be re-scheduled. Rather, after
    suspending, the VM resources are reserved for later resuming.
    """

    RESUME = "RESUME"
    """
    The VM is resumed using the saved memory state from the checkpoint file,
    and the VM's disk image. The VM is either started immediately, or
    re-scheduled depending on how it was suspended.
    """

    CANCEL = "CANCEL"
    """
    The VM is forcibly shutdown, its memory state is deleted. If a persistent
    disk image was used, that disk image is transferred back to the front-end.
    Any non-persistent disk images are deleted.
    """

    SHUTDOWN = "SHUTDOWN"
    """
    The VM is gracefully shutdown by sending the ACPI signal. If the VM does
    not shutdown, then it is considered to still be running. If successfully,
    shutdown, its memory state is deleted. If a persistent disk image was used,
    that disk image is transferred back to the front-end. Any non-persistent
    disk images are deleted.
    """

    REBOOT = "REBOOT"
    """
    Introduced in OpenNebula v3.2.

    The VM is gracefully restarted by sending the ACPI signal.
    """

    DONE = "DONE"
    """
    The VM is forcibly shutdown, its memory state is deleted. If a persistent
    disk image was used, that disk image is transferred back to the front-end.
    Any non-persistent disk images are deleted.
    """


class OpenNebulaResponse(XmlResponse):
    """
    XmlResponse class for the OpenNebula.org driver.
    """

    def success(self):
        """
        Check if response has the appropriate HTTP response code to be a
        success.

        :rtype:  ``bool``
        :return: True is success, else False.
        """
        i = int(self.status)
        return 200 <= i <= 299

    def parse_error(self):
        """
        Check if response contains any errors.

        @raise: :class:`InvalidCredsError`

        :rtype:  :class:`ElementTree`
        :return: Contents of HTTP response body.
        """
        if int(self.status) == httplib.UNAUTHORIZED:
            raise InvalidCredsError(self.body)
        return self.body


class OpenNebulaConnection(ConnectionUserAndKey):
    """
    Connection class for the OpenNebula.org driver.
    with plain_auth support
    """

    host = API_HOST
    port = API_PORT
    secure = API_SECURE
    plain_auth = API_PLAIN_AUTH
    responseCls = OpenNebulaResponse

    def __init__(self, *args, **kwargs):
        if "plain_auth" in kwargs:
            self.plain_auth = kwargs.pop("plain_auth")
        super().__init__(*args, **kwargs)

    def add_default_headers(self, headers):
        """
        Add headers required by the OpenNebula.org OCCI interface.

        Includes adding Basic HTTP Authorization headers for authenticating
        against the OpenNebula.org OCCI interface.

        :type  headers: ``dict``
        :param headers: Dictionary containing HTTP headers.

        :rtype:  ``dict``
        :return: Dictionary containing updated headers.
        """
        if self.plain_auth:
            passwd = self.key
        else:
            passwd = hashlib.sha1(b(self.key)).hexdigest()
        headers["Authorization"] = "Basic %s" % b64encode(
            b("{}:{}".format(self.user_id, passwd))
        ).decode("utf-8")
        return headers


class OpenNebulaNodeSize(NodeSize):
    """
    NodeSize class for the OpenNebula.org driver.
    """

    def __init__(self, id, name, ram, disk, bandwidth, price, driver, cpu=None, vcpu=None):
        super().__init__(
            id=id,
            name=name,
            ram=ram,
            disk=disk,
            bandwidth=bandwidth,
            price=price,
            driver=driver,
        )
        self.cpu = cpu
        self.vcpu = vcpu

    def __repr__(self):
        return (
            "<OpenNebulaNodeSize: id=%s, name=%s, ram=%s, disk=%s, "
            "bandwidth=%s, price=%s, driver=%s, cpu=%s, vcpu=%s ...>"
        ) % (
            self.id,
            self.name,
            self.ram,
            self.disk,
            self.bandwidth,
            self.price,
            self.driver.name,
            self.cpu,
            self.vcpu,
        )


class OpenNebulaNetwork:
    """
    Provide a common interface for handling networks of all types.

    Network objects are analogous to physical switches connecting two or
    more physical nodes together. The Network object provides the interface in
    libcloud through which we can manipulate networks in different cloud
    providers in the same way. Network objects don't actually do much directly
    themselves, instead the network driver handles the connection to the
    network.

    You don't normally create a network object yourself; instead you use
    a driver and then have that create the network for you.

    >>> from libcloud.compute.drivers.dummy import DummyNodeDriver
    >>> driver = DummyNodeDriver()
    >>> network = driver.create_network()
    >>> network = driver.list_networks()[0]
    >>> network.name
    'dummy-1'
    """

    def __init__(self, id, name, address, size, driver, extra=None):
        self.id = str(id)
        self.name = name
        self.address = address
        self.size = size
        self.driver = driver
        self.uuid = self.get_uuid()
        self.extra = extra or {}

    def get_uuid(self):
        """
        Unique hash for this network.

        The hash is a function of an SHA1 hash of the network's ID and
        its driver which means that it should be unique between all
        networks. In some subclasses there is no ID available so the
        public IP address is used. This means that, unlike a properly
        done system UUID, the same UUID may mean a different system
        install at a different time

        >>> from libcloud.network.drivers.dummy import DummyNetworkDriver
        >>> driver = DummyNetworkDriver()
        >>> network = driver.create_network()
        >>> network.get_uuid()
        'd3748461511d8b9b0e0bfa0d4d3383a619a2bb9f'

        Note, for example, that this example will always produce the
        same UUID!

        :rtype:  ``str``
        :return: Unique identifier for this instance.
        """
        return hashlib.sha1(b("{}:{}".format(self.id, self.driver.type))).hexdigest()

    def __repr__(self):
        return (
            "<OpenNebulaNetwork: uuid=%s, name=%s, address=%s, size=%s, " "provider=%s ...>"
        ) % (self.uuid, self.name, self.address, self.size, self.driver.name)


class OpenNebulaNodeDriver(NodeDriver):
    """
    OpenNebula.org node driver.
    """

    connectionCls = OpenNebulaConnection
    name = "OpenNebula"
    website = "http://opennebula.org/"
    type = Provider.OPENNEBULA

    NODE_STATE_MAP = {
        "INIT": NodeState.PENDING,
        "PENDING": NodeState.PENDING,
        "HOLD": NodeState.PENDING,
        "ACTIVE": NodeState.RUNNING,
        "STOPPED": NodeState.TERMINATED,
        "SUSPENDED": NodeState.PENDING,
        "DONE": NodeState.TERMINATED,
        "FAILED": NodeState.TERMINATED,
    }

    def __new__(cls, key, secret=None, api_version=DEFAULT_API_VERSION, **kwargs):
        if cls is OpenNebulaNodeDriver:
            if api_version in ["1.4"]:
                cls = OpenNebula_1_4_NodeDriver
            elif api_version in ["2.0", "2.2"]:
                cls = OpenNebula_2_0_NodeDriver
            elif api_version in ["3.0"]:
                cls = OpenNebula_3_0_NodeDriver
            elif api_version in ["3.2"]:
                cls = OpenNebula_3_2_NodeDriver
            elif api_version in ["3.6"]:
                cls = OpenNebula_3_6_NodeDriver
            elif api_version in ["3.8"]:
                cls = OpenNebula_3_8_NodeDriver
                if "plain_auth" not in kwargs:
                    kwargs["plain_auth"] = cls.plain_auth
                else:
                    cls.plain_auth = kwargs["plain_auth"]
            else:
                raise NotImplementedError(
                    "No OpenNebulaNodeDriver found for API version %s" % (api_version)
                )
            return super().__new__(cls)

    def create_node(self, name, size, image, networks=None):
        """
        Create a new OpenNebula node.

        @inherits: :class:`NodeDriver.create_node`

        :keyword networks: List of virtual networks to which this node should
                           connect. (optional)
        :type    networks: :class:`OpenNebulaNetwork` or
            ``list`` of :class:`OpenNebulaNetwork`
        """
        compute = ET.Element("COMPUTE")

        name = ET.SubElement(compute, "NAME")
        name.text = name

        instance_type = ET.SubElement(compute, "INSTANCE_TYPE")
        instance_type.text = size.name

        storage = ET.SubElement(compute, "STORAGE")
        ET.SubElement(storage, "DISK", {"image": "%s" % (str(image.id))})

        if networks:
            if not isinstance(networks, list):
                networks = [networks]

            networkGroup = ET.SubElement(compute, "NETWORK")
            for network in networks:
                if network.address:
                    ET.SubElement(
                        networkGroup,
                        "NIC",
                        {"network": "%s" % (str(network.id)), "ip": network.address},
                    )
                else:
                    ET.SubElement(networkGroup, "NIC", {"network": "%s" % (str(network.id))})

        xml = ET.tostring(compute)
        node = self.connection.request("/compute", method="POST", data=xml).object

        return self._to_node(node)

    def destroy_node(self, node):
        url = "/compute/%s" % (str(node.id))
        resp = self.connection.request(url, method="DELETE")

        return resp.status == httplib.OK

    def list_nodes(self):
        return self._to_nodes(self.connection.request("/compute").object)

    def list_images(self, location=None):
        return self._to_images(self.connection.request("/storage").object)

    def list_sizes(self, location=None):
        """
        Return list of sizes on a provider.

        @inherits: :class:`NodeDriver.list_sizes`

        :return: List of compute node sizes supported by the cloud provider.
        :rtype:  ``list`` of :class:`OpenNebulaNodeSize`
        """
        return [
            NodeSize(
                id=1,
                name="small",
                ram=None,
                disk=None,
                bandwidth=None,
                price=None,
                driver=self,
            ),
            NodeSize(
                id=2,
                name="medium",
                ram=None,
                disk=None,
                bandwidth=None,
                price=None,
                driver=self,
            ),
            NodeSize(
                id=3,
                name="large",
                ram=None,
                disk=None,
                bandwidth=None,
                price=None,
                driver=self,
            ),
        ]

    def list_locations(self):
        return [NodeLocation(0, "", "", self)]

    def ex_list_networks(self, location=None):
        """
        List virtual networks on a provider.

        :param location: Location from which to request a list of virtual
                         networks. (optional)
        :type  location: :class:`NodeLocation`

        :return: List of virtual networks available to be connected to a
                 compute node.
        :rtype:  ``list`` of :class:`OpenNebulaNetwork`
        """
        return self._to_networks(self.connection.request("/network").object)

    def ex_node_action(self, node, action):
        """
        Build action representation and instruct node to commit action.

        Build action representation from the compute node ID, and the
        action which should be carried out on that compute node. Then
        instruct the node to carry out that action.

        :param node: Compute node instance.
        :type  node: :class:`Node`

        :param action: Action to be carried out on the compute node.
        :type  action: ``str``

        :return: False if an HTTP Bad Request is received, else, True is
                 returned.
        :rtype:  ``bool``
        """
        compute_node_id = str(node.id)

        compute = ET.Element("COMPUTE")

        compute_id = ET.SubElement(compute, "ID")
        compute_id.text = compute_node_id

        state = ET.SubElement(compute, "STATE")
        state.text = action

        xml = ET.tostring(compute)

        url = "/compute/%s" % compute_node_id
        resp = self.connection.request(url, method="PUT", data=xml)

        if resp.status == httplib.BAD_REQUEST:
            return False
        else:
            return True

    def _to_images(self, object):
        """
        Request a list of images and convert that list to a list of NodeImage
        objects.

        Request a list of images from the OpenNebula web interface, and
        issue a request to convert each XML object representation of an image
        to a NodeImage object.

        :rtype:  ``list`` of :class:`NodeImage`
        :return: List of images.
        """
        images = []
        for element in object.findall("DISK"):
            image_id = element.attrib["href"].partition("/storage/")[2]
            image = self.connection.request("/storage/%s" % (image_id)).object
            images.append(self._to_image(image))

        return images

    def _to_image(self, image):
        """
        Take XML object containing an image description and convert to
        NodeImage object.

        :type  image: :class:`ElementTree`
        :param image: XML representation of an image.

        :rtype:  :class:`NodeImage`
        :return: The newly extracted :class:`NodeImage`.
        """
        return NodeImage(
            id=image.findtext("ID"),
            name=image.findtext("NAME"),
            driver=self.connection.driver,
            extra={"size": image.findtext("SIZE"), "url": image.findtext("URL")},
        )

    def _to_networks(self, object):
        """
        Request a list of networks and convert that list to a list of
        OpenNebulaNetwork objects.

        Request a list of networks from the OpenNebula web interface, and
        issue a request to convert each XML object representation of a network
        to an OpenNebulaNetwork object.

        :rtype:  ``list`` of :class:`OpenNebulaNetwork`
        :return: List of virtual networks.
        """
        networks = []
        for element in object.findall("NETWORK"):
            network_id = element.attrib["href"].partition("/network/")[2]
            network_element = self.connection.request("/network/%s" % (network_id)).object
            networks.append(self._to_network(network_element))

        return networks

    def _to_network(self, element):
        """
        Take XML object containing a network description and convert to
        OpenNebulaNetwork object.

        Take XML representation containing a network description and
        convert to OpenNebulaNetwork object.

        :rtype:  :class:`OpenNebulaNetwork`
        :return: The newly extracted :class:`OpenNebulaNetwork`.
        """
        return OpenNebulaNetwork(
            id=element.findtext("ID"),
            name=element.findtext("NAME"),
            address=element.findtext("ADDRESS"),
            size=element.findtext("SIZE"),
            driver=self.connection.driver,
        )

    def _to_nodes(self, object):
        """
        Request a list of compute nodes and convert that list to a list of
        Node objects.

        Request a list of compute nodes from the OpenNebula web interface, and
        issue a request to convert each XML object representation of a node
        to a Node object.

        :rtype:  ``list`` of :class:`Node`
        :return: A list of compute nodes.
        """
        computes = []
        for element in object.findall("COMPUTE"):
            compute_id = element.attrib["href"].partition("/compute/")[2]
            compute = self.connection.request("/compute/%s" % (compute_id)).object
            computes.append(self._to_node(compute))

        return computes

    def _to_node(self, compute):
        """
        Take XML object containing a compute node description and convert to
        Node object.

        Take XML representation containing a compute node description and
        convert to Node object.

        :type  compute: :class:`ElementTree`
        :param compute: XML representation of a compute node.

        :rtype:  :class:`Node`
        :return: The newly extracted :class:`Node`.
        """
        try:
            state = self.NODE_STATE_MAP[compute.findtext("STATE").upper()]
        except KeyError:
            state = NodeState.UNKNOWN

        return Node(
            id=compute.findtext("ID"),
            name=compute.findtext("NAME"),
            state=state,
            public_ips=self._extract_networks(compute),
            private_ips=[],
            driver=self.connection.driver,
            image=self._extract_images(compute),
        )

    def _extract_networks(self, compute):
        """
        Extract networks from a compute node XML representation.

        Extract network descriptions from a compute node XML representation,
        converting each network to an OpenNebulaNetwork object.

        :type  compute: :class:`ElementTree`
        :param compute: XML representation of a compute node.

        :rtype:  ``list`` of :class:`OpenNebulaNetwork`s.
        :return: List of virtual networks attached to the compute node.
        """
        networks = list()

        network_list = compute.find("NETWORK")
        for element in network_list.findall("NIC"):
            networks.append(
                OpenNebulaNetwork(
                    id=element.attrib.get("network", None),
                    name=None,
                    address=element.attrib.get("ip", None),
                    size=1,
                    driver=self.connection.driver,
                )
            )

        return networks

    def _extract_images(self, compute):
        """
        Extract image disks from a compute node XML representation.

        Extract image disk descriptions from a compute node XML representation,
        converting the disks to an NodeImage object.

        :type  compute: :class:`ElementTree`
        :param compute: XML representation of a compute node.

        :rtype:  :class:`NodeImage`.
        :return: First disk attached to a compute node.
        """
        disks = list()

        disk_list = compute.find("STORAGE")
        if disk_list is not None:
            for element in disk_list.findall("DISK"):
                disks.append(
                    NodeImage(
                        id=element.attrib.get("image", None),
                        name=None,
                        driver=self.connection.driver,
                        extra={"dev": element.attrib.get("dev", None)},
                    )
                )

        # @TODO: Return all disks when the Node type accepts multiple
        # attached disks per node.
        if len(disks) > 0:
            return disks[0]
        else:
            return None


class OpenNebula_1_4_NodeDriver(OpenNebulaNodeDriver):
    """
    OpenNebula.org node driver for OpenNebula.org v1.4.
    """

    name = "OpenNebula (v1.4)"


class OpenNebula_2_0_NodeDriver(OpenNebulaNodeDriver):
    """
    OpenNebula.org node driver for OpenNebula.org v2.0 through OpenNebula.org
    v2.2.
    """

    name = "OpenNebula (v2.0 - v2.2)"

    def create_node(self, name, size, image, networks=None, context=None):
        """
        Create a new OpenNebula node.

        @inherits: :class:`NodeDriver.create_node`

        :keyword networks: List of virtual networks to which this node should
                           connect. (optional)
        :type    networks: :class:`OpenNebulaNetwork` or ``list``
                           of :class:`OpenNebulaNetwork`

        :keyword context: Custom (key, value) pairs to be injected into
                          compute node XML description. (optional)
        :type    context: ``dict``

        :return: Instance of a newly created node.
        :rtype:  :class:`Node`
        """
        compute = ET.Element("COMPUTE")

        name = ET.SubElement(compute, "NAME")
        name.text = name

        instance_type = ET.SubElement(compute, "INSTANCE_TYPE")
        instance_type.text = size.name

        disk = ET.SubElement(compute, "DISK")
        ET.SubElement(disk, "STORAGE", {"href": "/storage/%s" % (str(image.id))})

        if networks:
            if not isinstance(networks, list):
                networks = [networks]

            for network in networks:
                nic = ET.SubElement(compute, "NIC")
                ET.SubElement(nic, "NETWORK", {"href": "/network/%s" % (str(network.id))})
                if network.address:
                    ip_line = ET.SubElement(nic, "IP")
                    ip_line.text = network.address

        if context and isinstance(context, dict):
            contextGroup = ET.SubElement(compute, "CONTEXT")
            for key, value in list(context.items()):
                context = ET.SubElement(contextGroup, key.upper())
                context.text = value

        xml = ET.tostring(compute)
        node = self.connection.request("/compute", method="POST", data=xml).object

        return self._to_node(node)

    def destroy_node(self, node):
        url = "/compute/%s" % (str(node.id))
        resp = self.connection.request(url, method="DELETE")

        return resp.status == httplib.NO_CONTENT

    def list_sizes(self, location=None):
        """
        Return list of sizes on a provider.

        @inherits: :class:`NodeDriver.list_sizes`

        :return: List of compute node sizes supported by the cloud provider.
        :rtype:  ``list`` of :class:`OpenNebulaNodeSize`
        """
        return [
            OpenNebulaNodeSize(
                id=1,
                name="small",
                ram=1024,
                cpu=1,
                disk=None,
                bandwidth=None,
                price=None,
                driver=self,
            ),
            OpenNebulaNodeSize(
                id=2,
                name="medium",
                ram=4096,
                cpu=4,
                disk=None,
                bandwidth=None,
                price=None,
                driver=self,
            ),
            OpenNebulaNodeSize(
                id=3,
                name="large",
                ram=8192,
                cpu=8,
                disk=None,
                bandwidth=None,
                price=None,
                driver=self,
            ),
            OpenNebulaNodeSize(
                id=4,
                name="custom",
                ram=0,
                cpu=0,
                disk=None,
                bandwidth=None,
                price=None,
                driver=self,
            ),
        ]

    def _to_images(self, object):
        """
        Request a list of images and convert that list to a list of NodeImage
        objects.

        Request a list of images from the OpenNebula web interface, and
        issue a request to convert each XML object representation of an image
        to a NodeImage object.

        :rtype:  ``list`` of :class:`NodeImage`
        :return: List of images.
        """
        images = []
        for element in object.findall("STORAGE"):
            image_id = element.attrib["href"].partition("/storage/")[2]
            image = self.connection.request("/storage/%s" % (image_id)).object
            images.append(self._to_image(image))

        return images

    def _to_image(self, image):
        """
        Take XML object containing an image description and convert to
        NodeImage object.

        :type  image: :class:`ElementTree`
        :param image: XML representation of an image.

        :rtype:  :class:`NodeImage`
        :return: The newly extracted :class:`NodeImage`.
        """
        return NodeImage(
            id=image.findtext("ID"),
            name=image.findtext("NAME"),
            driver=self.connection.driver,
            extra={
                "description": image.findtext("DESCRIPTION"),
                "type": image.findtext("TYPE"),
                "size": image.findtext("SIZE"),
                "fstype": image.findtext("FSTYPE", None),
            },
        )

    def _to_node(self, compute):
        """
        Take XML object containing a compute node description and convert to
        Node object.

        Take XML representation containing a compute node description and
        convert to Node object.

        :type  compute: :class:`ElementTree`
        :param compute: XML representation of a compute node.

        :rtype:  :class:`Node`
        :return: The newly extracted :class:`Node`.
        """
        try:
            state = self.NODE_STATE_MAP[compute.findtext("STATE").upper()]
        except KeyError:
            state = NodeState.UNKNOWN

        return Node(
            id=compute.findtext("ID"),
            name=compute.findtext("NAME"),
            state=state,
            public_ips=self._extract_networks(compute),
            private_ips=[],
            driver=self.connection.driver,
            image=self._extract_images(compute),
            size=self._extract_size(compute),
            extra={"context": self._extract_context(compute)},
        )

    def _extract_networks(self, compute):
        """
        Extract networks from a compute node XML representation.

        Extract network descriptions from a compute node XML representation,
        converting each network to an OpenNebulaNetwork object.

        :type  compute: :class:`ElementTree`
        :param compute: XML representation of a compute node.

        :rtype:  ``list`` of :class:`OpenNebulaNetwork`
        :return: List of virtual networks attached to the compute node.
        """
        networks = []

        for element in compute.findall("NIC"):
            network = element.find("NETWORK")
            network_id = network.attrib["href"].partition("/network/")[2]

            networks.append(
                OpenNebulaNetwork(
                    id=network_id,
                    name=network.attrib.get("name", None),
                    address=element.findtext("IP"),
                    size=1,
                    driver=self.connection.driver,
                    extra={"mac": element.findtext("MAC")},
                )
            )

        return networks

    def _extract_images(self, compute):
        """
        Extract image disks from a compute node XML representation.

        Extract image disk descriptions from a compute node XML representation,
        converting the disks to an NodeImage object.

        :type  compute: :class:`ElementTree`
        :param compute: XML representation of a compute node.

        :rtype:  ``list`` of :class:`NodeImage`
        :return: Disks attached to a compute node.
        """
        disks = list()

        for element in compute.findall("DISK"):
            disk = element.find("STORAGE")
            image_id = disk.attrib["href"].partition("/storage/")[2]

            if "id" in element.attrib:
                disk_id = element.attrib["id"]
            else:
                disk_id = None

            disks.append(
                NodeImage(
                    id=image_id,
                    name=disk.attrib.get("name", None),
                    driver=self.connection.driver,
                    extra={
                        "type": element.findtext("TYPE"),
                        "disk_id": disk_id,
                        "target": element.findtext("TARGET"),
                    },
                )
            )

        # Return all disks when the Node type accepts multiple attached disks
        # per node.
        if len(disks) > 1:
            return disks
        elif len(disks) == 1:
            return disks[0]
        else:
            return None

    def _extract_size(self, compute):
        """
        Extract size, or node type, from a compute node XML representation.

        Extract node size, or node type, description from a compute node XML
        representation, converting the node size to a NodeSize object.

        :type  compute: :class:`ElementTree`
        :param compute: XML representation of a compute node.

        :rtype:  :class:`OpenNebulaNodeSize`
        :return: Node type of compute node.
        """
        instance_type = compute.find("INSTANCE_TYPE")

        try:
            return next(
                node_size for node_size in self.list_sizes() if node_size.name == instance_type.text
            )
        except StopIteration:
            return None

    def _extract_context(self, compute):
        """
        Extract size, or node type, from a compute node XML representation.

        Extract node size, or node type, description from a compute node XML
        representation, converting the node size to a NodeSize object.

        :type  compute: :class:`ElementTree`
        :param compute: XML representation of a compute node.

        :rtype:  ``dict``
        :return: Dictionary containing (key, value) pairs related to
                 compute node context.
        """
        contexts = dict()
        context = compute.find("CONTEXT")

        if context is not None:
            for context_element in list(context):
                contexts[context_element.tag.lower()] = context_element.text

        return contexts


class OpenNebula_3_0_NodeDriver(OpenNebula_2_0_NodeDriver):
    """
    OpenNebula.org node driver for OpenNebula.org v3.0.
    """

    name = "OpenNebula (v3.0)"

    def ex_node_set_save_name(self, node, name):
        """
        Build action representation and instruct node to commit action.

        Build action representation from the compute node ID, the disk image
        which will be saved, and the name under which the image will be saved
        upon shutting down the compute node.

        :param node: Compute node instance.
        :type  node: :class:`Node`

        :param name: Name under which the image should be saved after shutting
                     down the compute node.
        :type  name: ``str``

        :return: False if an HTTP Bad Request is received, else, True is
                 returned.
        :rtype:  ``bool``
        """
        compute_node_id = str(node.id)

        compute = ET.Element("COMPUTE")

        compute_id = ET.SubElement(compute, "ID")
        compute_id.text = compute_node_id

        disk = ET.SubElement(compute, "DISK", {"id": str(node.image.id)})

        ET.SubElement(
            disk,
            "STORAGE",
            {"href": "/storage/%s" % (str(node.image.id)), "name": node.image.name},
        )

        ET.SubElement(disk, "SAVE_AS", {"name": str(name)})

        xml = ET.tostring(compute)

        url = "/compute/%s" % compute_node_id
        resp = self.connection.request(url, method="PUT", data=xml)

        if resp.status == httplib.BAD_REQUEST:
            return False
        else:
            return True

    def _to_network(self, element):
        """
        Take XML object containing a network description and convert to
        OpenNebulaNetwork object.

        Take XML representation containing a network description and
        convert to OpenNebulaNetwork object.

        :return: The newly extracted :class:`OpenNebulaNetwork`.
        :rtype:  :class:`OpenNebulaNetwork`
        """
        return OpenNebulaNetwork(
            id=element.findtext("ID"),
            name=element.findtext("NAME"),
            address=element.findtext("ADDRESS"),
            size=element.findtext("SIZE"),
            driver=self.connection.driver,
            extra={"public": element.findtext("PUBLIC")},
        )


class OpenNebula_3_2_NodeDriver(OpenNebula_3_0_NodeDriver):
    """
    OpenNebula.org node driver for OpenNebula.org v3.2.
    """

    name = "OpenNebula (v3.2)"

    def reboot_node(self, node):
        return self.ex_node_action(node, ACTION.REBOOT)

    def list_sizes(self, location=None):
        """
        Return list of sizes on a provider.

        @inherits: :class:`NodeDriver.list_sizes`

        :return: List of compute node sizes supported by the cloud provider.
        :rtype:  ``list`` of :class:`OpenNebulaNodeSize`
        """
        return self._to_sizes(self.connection.request("/instance_type").object)

    def _to_sizes(self, object):
        """
        Request a list of instance types and convert that list to a list of
        OpenNebulaNodeSize objects.

        Request a list of instance types from the OpenNebula web interface,
        and issue a request to convert each XML object representation of an
        instance type to an OpenNebulaNodeSize object.

        :return: List of instance types.
        :rtype:  ``list`` of :class:`OpenNebulaNodeSize`
        """
        sizes = []
        size_id = 1

        attributes = [
            ("name", str, None),
            ("ram", int, "MEMORY"),
            ("cpu", float, None),
            ("vcpu", float, None),
            ("disk", str, None),
            ("bandwidth", float, None),
            ("price", float, None),
        ]

        for element in object.findall("INSTANCE_TYPE"):
            size_kwargs = {"id": size_id, "driver": self}
            values = self._get_attributes_values(attributes=attributes, element=element)
            size_kwargs.update(values)

            size = OpenNebulaNodeSize(**size_kwargs)
            sizes.append(size)
            size_id += 1

        return sizes

    def _get_attributes_values(self, attributes, element):
        values = {}

        for attribute_name, attribute_type, alias in attributes:
            key = alias if alias else attribute_name.upper()
            value = element.findtext(key)

            if value is not None:
                value = attribute_type(value)

            values[attribute_name] = value

        return values


class OpenNebula_3_6_NodeDriver(OpenNebula_3_2_NodeDriver):
    """
    OpenNebula.org node driver for OpenNebula.org v3.6.
    """

    name = "OpenNebula (v3.6)"

    def create_volume(self, size, name, location=None, snapshot=None):
        storage = ET.Element("STORAGE")

        vol_name = ET.SubElement(storage, "NAME")
        vol_name.text = name

        vol_type = ET.SubElement(storage, "TYPE")
        vol_type.text = "DATABLOCK"

        description = ET.SubElement(storage, "DESCRIPTION")
        description.text = "Attached storage"

        public = ET.SubElement(storage, "PUBLIC")
        public.text = "NO"

        persistent = ET.SubElement(storage, "PERSISTENT")
        persistent.text = "YES"

        fstype = ET.SubElement(storage, "FSTYPE")
        fstype.text = "ext3"

        vol_size = ET.SubElement(storage, "SIZE")
        vol_size.text = str(size)

        xml = ET.tostring(storage)
        volume = self.connection.request("/storage", {"occixml": xml}, method="POST").object

        return self._to_volume(volume)

    def destroy_volume(self, volume):
        url = "/storage/%s" % (str(volume.id))
        resp = self.connection.request(url, method="DELETE")

        return resp.status == httplib.NO_CONTENT

    def attach_volume(self, node, volume, device):
        action = ET.Element("ACTION")

        perform = ET.SubElement(action, "PERFORM")
        perform.text = "ATTACHDISK"

        params = ET.SubElement(action, "PARAMS")

        ET.SubElement(params, "STORAGE", {"href": "/storage/%s" % (str(volume.id))})

        target = ET.SubElement(params, "TARGET")
        target.text = device

        xml = ET.tostring(action)

        url = "/compute/%s/action" % node.id

        resp = self.connection.request(url, method="POST", data=xml)
        return resp.status == httplib.ACCEPTED

    def _do_detach_volume(self, node_id, disk_id):
        action = ET.Element("ACTION")

        perform = ET.SubElement(action, "PERFORM")
        perform.text = "DETACHDISK"

        params = ET.SubElement(action, "PARAMS")

        ET.SubElement(params, "DISK", {"id": disk_id})

        xml = ET.tostring(action)

        url = "/compute/%s/action" % node_id

        resp = self.connection.request(url, method="POST", data=xml)
        return resp.status == httplib.ACCEPTED

    def detach_volume(self, volume):
        # We need to find the node using this volume
        for node in self.list_nodes():
            if type(node.image) is not list:
                # This node has only one associated image. It is not the one we
                # are after.
                continue

            for disk in node.image:
                if disk.id == volume.id:
                    # Node found. We can now detach the volume
                    disk_id = disk.extra["disk_id"]
                    return self._do_detach_volume(node.id, disk_id)

        return False

    def list_volumes(self):
        return self._to_volumes(self.connection.request("/storage").object)

    def _to_volume(self, storage):
        return StorageVolume(
            id=storage.findtext("ID"),
            name=storage.findtext("NAME"),
            size=int(storage.findtext("SIZE")),
            driver=self.connection.driver,
        )

    def _to_volumes(self, object):
        volumes = []
        for storage in object.findall("STORAGE"):
            storage_id = storage.attrib["href"].partition("/storage/")[2]

            volumes.append(
                self._to_volume(self.connection.request("/storage/%s" % storage_id).object)
            )

        return volumes


class OpenNebula_3_8_NodeDriver(OpenNebula_3_6_NodeDriver):
    """
    OpenNebula.org node driver for OpenNebula.org v3.8.
    """

    name = "OpenNebula (v3.8)"
    plain_auth = API_PLAIN_AUTH

    def _to_sizes(self, object):
        """
        Request a list of instance types and convert that list to a list of
        OpenNebulaNodeSize objects.

        Request a list of instance types from the OpenNebula web interface,
        and issue a request to convert each XML object representation of an
        instance type to an OpenNebulaNodeSize object.

        :return: List of instance types.
        :rtype:  ``list`` of :class:`OpenNebulaNodeSize`
        """
        sizes = []
        size_id = 1

        attributes = [
            ("name", str, None),
            ("ram", int, "MEMORY"),
            ("cpu", float, None),
            ("vcpu", float, None),
            ("disk", str, None),
            ("bandwidth", float, None),
            ("price", float, None),
        ]

        for element in object.findall("INSTANCE_TYPE"):
            element = self.connection.request(
                ("/instance_type/%s") % (element.attrib["name"])
            ).object

            size_kwargs = {"id": size_id, "driver": self}
            values = self._get_attributes_values(attributes=attributes, element=element)
            size_kwargs.update(values)

            size = OpenNebulaNodeSize(**size_kwargs)
            sizes.append(size)
            size_id += 1
        return sizes

    def _ex_connection_class_kwargs(self):
        """
        Set plain_auth as an extra :class:`OpenNebulaConnection_3_8` argument

        :return: ``dict`` of :class:`OpenNebulaConnection_3_8` input arguments
        """

        return {"plain_auth": self.plain_auth}
